==> fred_frgal-3.15.0.tar.gz: fred_frgal-3.15.0/frgal/utils.py <==
"""Various utilities."""
import asyncio
import random
import string
import warnings
from functools import partial
from threading import Thread
from typing import Any, AsyncIterable, Awaitable, Callable, ClassVar, Iterable, List, Optional, Sequence, Tuple, TypeVar
from unittest.mock import AsyncMock as UnittestAsyncMock, Mock, _Call

import grpc
from asyncstdlib import list as alist

from ._utils import MESSAGE_MAX_LENGTH as _MESSAGE_MAX_LENGTH, format_message as _format_message  # noqa: F401


def __getattr__(name: str) -> Any:
    if name in ('MESSAGE_MAX_LENGTH', 'format_message'):  # pragma: nocover
        warnings.warn(f"'{name}' is deprecated and will be removed from public API.", DeprecationWarning, stacklevel=2)
        return globals()["_" + name]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


T = TypeVar('T')


def make_awaitable(value: T) -> Awaitable[T]:
    """Convert value to an awaitable."""
    async def _wrapper() -> T:
        return value
    return _wrapper()


def get_call_id() -> str:
    """Return random call ID for logging."""
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))  # nosec


class CallableMock:
    """Mock of a gRPC method Callable object."""

    def __init__(self, method: str, call: Callable, *, exhaust_iterables: Optional[bool] = None):
        self._method = method.encode('utf-8')
        self.call = call
        self.exhaust_iterables = exhaust_iterables

    def __call__(self, request: Any, *args: Any, **kwargs: Any) -> Any:
        """Mock gRPC method call."""
        if self.exhaust_iterables:
            request = self._exhaust(request)
        elif (self.exhaust_iterables is None and not isinstance(request, Sequence)
              and isinstance(request, (Iterable, AsyncIterable))):
            warnings.warn(
                'Default value of `exhaust_iterables` will change to True in next major release.'
                'If you want to maintain the current behaviour, set it to False explicitely.',
                DeprecationWarning,
            )

        return self.call(request, *args, **kwargs)

    def future(self, *args: Any, **kwargs: Any) -> Mock:
        """Mock asynchronous gRPC method call."""
        future_mock = Mock(name='future', spec=('result', ))
        future_mock.result.return_value = self.call(*args, **kwargs)
        return future_mock

    @staticmethod
    def _exhaust(value: Any) -> Any:
        """Exhaust any iterables."""
        results: List[Any] = []
        exceptions: List[Exception] = []

        def _inner(value: AsyncIterable, results: List[Any], exceptions: List[Exception]) -> Any:
            try:
                results.append(tuple(asyncio.run(alist(value))))
            except Exception as err:
                exceptions.append(err)

        if isinstance(value, Sequence):
            # Strings, lists and tuples should be left untouched
            return value
        elif isinstance(value, Iterable):
            return tuple(value)
        elif isinstance(value, AsyncIterable):
            # Run async iterable in separate thead – this works both
            # with and without running async loop in the main thread
            thread = Thread(target=_inner, args=(value, results, exceptions))
            thread.start()
            thread.join()
            if exceptions:
                raise exceptions[0]
            else:
                return results[0]
        else:
            return value


class AsyncMock(Mock):
    """Async mock variant."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Async mock call."""
        result = super().__call__(*args, **kwargs)
        if isinstance(result, (AsyncIterable, Awaitable)):
            return result
        else:
            warnings.warn(
                'Automagical transformation to awaitables is deprecated '
                'and will be removed in the next major release.',
                DeprecationWarning,
            )
            return make_awaitable(result)


class MockingChannel(grpc.Channel):
    """A channel which binds Stub calls to TestClient internal mock."""

    def __init__(self, client: 'TestClientMixin', exhaust_iterables: Optional[bool] = None):
        self.client = client
        self.exhaust_iterables = exhaust_iterables

    def subscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unsubscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unary_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def unary_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def close(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""


class TestClientMixin:
    """Mixin to wrap gRPC client for tests.

    Attributes:
        mock: A mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    exhaust_iterables: ClassVar[Optional[bool]] = None

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.mock = Mock(spec=('__call__', ))

    def connect(self) -> None:
        """Create a mocking channel which redirects all gRPC method calls to the mock attribute."""
        self.channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)


class AsyncTestClientMixin(TestClientMixin):
    """Mixin to wrap async gRPC client for tests.

    Attributes:
        mock: An async mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.mock = AsyncMock(spec=('__call__', ))
        self._channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)  # type: Any
        self._channel._loop = asyncio.new_event_loop()

    def __del__(self) -> None:
        self._channel._loop.close()


async def _consume_async_gen(gen: Callable[..., AsyncIterable], *args: Any, **kwargs: Any) -> list:
    return await alist(gen(*args, **kwargs))


def get_mock_awaits(mock: UnittestAsyncMock) -> List[_Call]:
    """Return a list of awaited call from mock and all its children.

    Awaits are in an order of their coroutine calls.
    """
    # A list of awaited calls.
    awaits = []
    # A queue of mock objects to process.
    queue: List[Tuple[str, Mock]] = []
    queue.append(('', mock))
    # Collect awaits from the structure of mocks.
    # Based on `CallableMixin._increment_mock_call`.
    while queue:
        parent_name, a_mock = queue.pop(0)

        # Collect children to process.
        children = []
        if isinstance(a_mock._mock_return_value, Mock):
            children.append(a_mock._mock_return_value)
        children.extend(a_mock._mock_children.values())

        # Enqueue children.
        for child in children:
            # Create a new parent name for the child.
            if child._mock_new_name == '()':
                dot = ''
            else:
                dot = '.'
            if parent_name:
                new_name = parent_name + dot + child._mock_new_name
            else:
                new_name = child._mock_new_name

            queue.append((new_name, child))

        # Collect calls for the current mock.
        for call in a_mock._mock_await_args_list:
            new_name = parent_name + (call._mock_name or '')
            new_call = _Call((new_name, ) + call._get_call_arguments())
            awaits.append(new_call)

    result = []
    for call in mock.mock_calls:
        if call in awaits:
            awaits.remove(call)
            result.append(call)
    return result

==> fred_frgal-3.15.1.tar.gz: fred_frgal-3.15.1/frgal/utils.py <==
"""Various utilities."""
import asyncio
import warnings
from functools import partial
from threading import Thread
from typing import Any, AsyncIterable, Awaitable, Callable, ClassVar, Iterable, List, Optional, Sequence, Tuple, TypeVar
from unittest.mock import AsyncMock as UnittestAsyncMock, Mock, _Call

import grpc
from asyncstdlib import list as alist

from ._utils import (MESSAGE_MAX_LENGTH as _MESSAGE_MAX_LENGTH, format_message as _format_message,  # noqa: F401
                     get_call_id as _get_call_id)


def __getattr__(name: str) -> Any:
    if name in ('MESSAGE_MAX_LENGTH', 'format_message', 'get_call_id'):  # pragma: nocover
        warnings.warn(f"'{name}' is deprecated and will be removed from public API.", DeprecationWarning, stacklevel=2)
        return globals()["_" + name]
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


T = TypeVar('T')


def make_awaitable(value: T) -> Awaitable[T]:
    """Convert value to an awaitable."""
    async def _wrapper() -> T:
        return value
    return _wrapper()


class CallableMock:
    """Mock of a gRPC method Callable object."""

    def __init__(self, method: str, call: Callable, *, exhaust_iterables: Optional[bool] = None):
        self._method = method.encode('utf-8')
        self.call = call
        self.exhaust_iterables = exhaust_iterables

    def __call__(self, request: Any, *args: Any, **kwargs: Any) -> Any:
        """Mock gRPC method call."""
        if self.exhaust_iterables:
            request = self._exhaust(request)
        elif (self.exhaust_iterables is None and not isinstance(request, Sequence)
              and isinstance(request, (Iterable, AsyncIterable))):
            warnings.warn(
                'Default value of `exhaust_iterables` will change to True in next major release.'
                'If you want to maintain the current behaviour, set it to False explicitely.',
                DeprecationWarning,
            )

        return self.call(request, *args, **kwargs)

    def future(self, *args: Any, **kwargs: Any) -> Mock:
        """Mock asynchronous gRPC method call."""
        future_mock = Mock(name='future', spec=('result', ))
        future_mock.result.return_value = self.call(*args, **kwargs)
        return future_mock

    @staticmethod
    def _exhaust(value: Any) -> Any:
        """Exhaust any iterables."""
        results: List[Any] = []
        exceptions: List[Exception] = []

        def _inner(value: AsyncIterable, results: List[Any], exceptions: List[Exception]) -> Any:
            try:
                results.append(tuple(asyncio.run(alist(value))))
            except Exception as err:
                exceptions.append(err)

        if isinstance(value, Sequence):
            # Strings, lists and tuples should be left untouched
            return value
        elif isinstance(value, Iterable):
            return tuple(value)
        elif isinstance(value, AsyncIterable):
            # Run async iterable in separate thead – this works both
            # with and without running async loop in the main thread
            thread = Thread(target=_inner, args=(value, results, exceptions))
            thread.start()
            thread.join()
            if exceptions:
                raise exceptions[0]
            else:
                return results[0]
        else:
            return value


class AsyncMock(Mock):
    """Async mock variant."""

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Async mock call."""
        result = super().__call__(*args, **kwargs)
        if isinstance(result, (AsyncIterable, Awaitable)):
            return result
        else:
            warnings.warn(
                'Automagical transformation to awaitables is deprecated '
                'and will be removed in the next major release.',
                DeprecationWarning,
            )
            return make_awaitable(result)


class MockingChannel(grpc.Channel):
    """A channel which binds Stub calls to TestClient internal mock."""

    def __init__(self, client: 'TestClientMixin', exhaust_iterables: Optional[bool] = None):
        self.client = client
        self.exhaust_iterables = exhaust_iterables

    def subscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unsubscribe(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""

    def unary_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def unary_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_unary(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def stream_stream(self, method: str, *args: Any, **kwargs: Any) -> CallableMock:
        """Return a mocked method bound to the test client mock."""
        return CallableMock(method, partial(self.client.mock, method=method), exhaust_iterables=self.exhaust_iterables)

    def close(self, *args: Any, **kwargs: Any) -> None:
        """Noop."""


class TestClientMixin:
    """Mixin to wrap gRPC client for tests.

    Attributes:
        mock: A mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    exhaust_iterables: ClassVar[Optional[bool]] = None

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.mock = Mock(spec=('__call__', ))

    def connect(self) -> None:
        """Create a mocking channel which redirects all gRPC method calls to the mock attribute."""
        self.channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)


class AsyncTestClientMixin(TestClientMixin):
    """Mixin to wrap async gRPC client for tests.

    Attributes:
        mock: An async mock which handles all gRPC method calls.
              Acts with a signature `(request: grpc.Message, method: str, timeout: int = None)`.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.mock = AsyncMock(spec=('__call__', ))
        self._channel = MockingChannel(self, exhaust_iterables=self.exhaust_iterables)  # type: Any
        self._channel._loop = asyncio.new_event_loop()

    def __del__(self) -> None:
        self._channel._loop.close()


async def _consume_async_gen(gen: Callable[..., AsyncIterable], *args: Any, **kwargs: Any) -> list:
    return await alist(gen(*args, **kwargs))


def get_mock_awaits(mock: UnittestAsyncMock) -> List[_Call]:
    """Return a list of awaited call from mock and all its children.

    Awaits are in an order of their coroutine calls.
    """
    # A list of awaited calls.
    awaits = []
    # A queue of mock objects to process.
    queue: List[Tuple[str, Mock]] = []
    queue.append(('', mock))
    # Collect awaits from the structure of mocks.
    # Based on `CallableMixin._increment_mock_call`.
    while queue:
        parent_name, a_mock = queue.pop(0)

        # Collect children to process.
        children = []
        if isinstance(a_mock._mock_return_value, Mock):
            children.append(a_mock._mock_return_value)
        children.extend(a_mock._mock_children.values())

        # Enqueue children.
        for child in children:
            # Create a new parent name for the child.
            if child._mock_new_name == '()':
                dot = ''
            else:
                dot = '.'
            if parent_name:
                new_name = parent_name + dot + child._mock_new_name
            else:
                new_name = child._mock_new_name

            queue.append((new_name, child))

        # Collect calls for the current mock.
        for call in a_mock._mock_await_args_list:
            new_name = parent_name + (call._mock_name or '')
            new_call = _Call((new_name, ) + call._get_call_arguments())
            awaits.append(new_call)

    result = []
    for call in mock.mock_calls:
        if call in awaits:
            awaits.remove(call)
            result.append(call)
    return result
